@@ -15,64 +15,36 @@ The :class:`APIManager` class allow users to create ReSTful APIs for
 their SQLAlchemy models.
 
 """
+import warnings
 from collections import defaultdict
-from collections import namedtuple
+from typing import Dict
+from typing import Optional
 from uuid import uuid1
 
 from flask import Blueprint
 
-from .helpers import collection_name
-from .helpers import model_for
-from .helpers import primary_key_for
+from . import registry
+from .helpers import get_model
 from .helpers import primary_key_names
-from .helpers import serializer_for
-from .helpers import url_for
 from .serialization import DefaultDeserializer
-from .serialization import FastSerializer
+from .serialization import DefaultSerializer
+from .serialization import Deserializer
+from .serialization import Serializer
 from .views import API
-from .views import FunctionAPI
 from .views import RelationshipAPI
+from .views.base import FetchCollection
+from .views.base import FetchResource
+from .views.function import FunctionAPI
 
 #: The names of HTTP methods that allow fetching information.
 READONLY_METHODS = frozenset(('GET', ))
 
-#: The names of HTTP methods that allow creating, updating, or deleting
-#: information.
+#: The names of HTTP methods that allow creating, updating, or deleting information.
 WRITEONLY_METHODS = frozenset(('PATCH', 'POST', 'DELETE'))
 
 #: The set of all recognized HTTP methods.
 ALL_METHODS = READONLY_METHODS | WRITEONLY_METHODS
 
-#: The default URL prefix for APIs created by instance of :class:`APIManager`.
-DEFAULT_URL_PREFIX = '/api'
-
-STRING_TYPES = (str, )
-
-#: A triple that stores the SQLAlchemy session and the universal pre- and post-
-#: processors to be applied to any API created for a particular Flask
-#: application.
-#:
-#: These tuples are used by :class:`APIManager` to store information about
-#: Flask applications registered using :meth:`APIManager.init_app`.
-# RestlessInfo = namedtuple('RestlessInfo', ['session',
-#                                            'universal_preprocessors',
-#                                            'universal_postprocessors'])
-
-#: A tuple that stores information about a created API.
-#:
-#: The elements are, in order,
-#:
-#: - `collection_name`, the name by which a collection of instances of
-#:   the model exposed by this API is known,
-#: - `blueprint_name`, the name of the blueprint that contains this API,
-#: - `serializer`, the subclass of :class:`Serializer` provided for the
-#:   model exposed by this API.
-#: - `primary_key`, the primary key used by the model
-#: - `url_prefix`, the url prefix to use for the collection
-#:
-APIInfo = namedtuple('APIInfo', ['collection_name', 'blueprint_name', 'serializer',
-                                 'primary_key', 'url_prefix'])
-
 
 class IllegalArgumentError(Exception):
     """This exception is raised when a calling function has provided illegal
@@ -82,7 +54,7 @@ class IllegalArgumentError(Exception):
     pass
 
 
-class APIManager(object):
+class APIManager:
     """Provides a method for creating a public ReSTful JSON API with respect
     to a given :class:`~flask.Flask` application object.
 
@@ -96,12 +68,6 @@ class APIManager(object):
     `session` is the :class:`~sqlalchemy.orm.session.Session` object in
     which changes to the database will be made.
 
-    `flask_sqlalchemy_db` is the :class:`~flask.ext.sqlalchemy.SQLAlchemy`
-    object with which `app` has been registered and which contains the
-    database models for which API endpoints will be created.
-
-    If `flask_sqlalchemy_db` is not ``None``, `session` will be ignored.
-
     For example, to use this class with models defined in pure SQLAlchemy::
 
         from flask import Flask
@@ -111,19 +77,9 @@ class APIManager(object):
 
         engine = create_engine('sqlite:////tmp/mydb.sqlite')
         Session = sessionmaker(bind=engine)
-        mysession = Session()
+        my_session = Session()
         app = Flask(__name__)
-        apimanager = APIManager(app, session=mysession)
-
-    and with models defined with Flask-SQLAlchemy::
-
-        from flask import Flask
-        from flask.ext.restless import APIManager
-        from flask.ext.sqlalchemy import SQLAlchemy
-
-        app = Flask(__name__)
-        db = SQLALchemy(app)
-        apimanager = APIManager(app, flask_sqlalchemy_db=db)
+        api_manager = APIManager(app, session=my_session)
 
     `url_prefix` is the URL prefix at which each API created by this
     instance will be accessible. For example, if this is set to
@@ -144,104 +100,38 @@ class APIManager(object):
     information on using preprocessors and postprocessors, see
     :ref:`processors`.
 
+    `include_links` controls whether to include link objects in resource objects
+    https://jsonapi.org/format/#document-links
+
     """
 
-    #: The format of the name of the API view for a given model.
-    #:
-    #: This format string expects the name of a model to be provided when
-    #: formatting.
-    APINAME_FORMAT = '{0}api'
-
-    def __init__(self, app=None, session=None, flask_sqlalchemy_db=None,
-                 preprocessors=None, postprocessors=None, url_prefix=None):
-        if session is None and flask_sqlalchemy_db is None:
-            msg = 'must specify either `flask_sqlalchemy_db` or `session`'
-            raise ValueError(msg)
+    def __init__(self, app=None, session=None, preprocessors=None, postprocessors=None, url_prefix='/api', include_links: bool = False):
+        if session is None:
+            raise ValueError('`session` can not be empty')
 
         self.app = app
 
-        # Stash this instance so that it can be examined later by the global
-        # `url_for`, `model_for`, and `collection_name` functions.
-        #
-        # TODO This is a bit of poor code style because it requires the
-        # APIManager to know about these global functions that use it.
-        url_for.register(self)
-        model_for.register(self)
-        collection_name.register(self)
-        serializer_for.register(self)
-        primary_key_for.register(self)
-
         #: A mapping whose keys are models for which this object has
         #: created an API via the :meth:`create_api_blueprint` method
         #: and whose values are the corresponding collection names for
         #: those models.
-        self.created_apis_for = {}
+        self.created_apis_for: dict = {}
 
         #: List of blueprints created by :meth:`create_api` to be registered
         #: to the app when calling :meth:`init_app`.
-        self.blueprints = []
+        self.blueprints: list = []
 
-        # If a Flask-SQLAlchemy object is provided, prefer the session
-        # from that object.
-        if flask_sqlalchemy_db is not None:
-            session = flask_sqlalchemy_db.session
-
-        # pre = preprocessors or {}
-        # post = postprocessors or {}
-        # self.restless_info = RestlessInfo(session, pre, post)
         self.pre = preprocessors or {}
         self.post = postprocessors or {}
         self.session = session
 
         #: The default URL prefix for APIs created by this manager.
         #:
-        #: This can be overriden by the `url_prefix` keyword argument in the
+        #: This can be overridden by the `url_prefix` keyword argument in the
         #: :meth:`create_api` method.
         self.url_prefix = url_prefix
 
-        # if self.app is not None:
-        #     self.init_app(self.app)
-
-    @staticmethod
-    def api_name(collection_name):
-        """Returns the name of the :class:`API` instance exposing models of the
-        specified type of collection.
-
-        `collection_name` must be a string.
-
-        """
-        return APIManager.APINAME_FORMAT.format(collection_name)
-
-    def model_for(self, collection_name):
-        """Returns the SQLAlchemy model class whose type is given by the
-        specified collection name.
-
-        `collection_name` is a string containing the collection name as
-        provided to the ``collection_name`` keyword argument to
-        :meth:`create_api_blueprint`.
-
-        The collection name should correspond to a model on which
-        :meth:`create_api_blueprint` has been invoked previously. If it doesn't
-        this method raises :exc:`ValueError`.
-
-        This method is the inverse of :meth:`collection_name`::
-
-            >>> from mymodels import Person
-            >>> manager.create_api(Person, collection_name='people')
-            >>> manager.collection_name(manager.model_for('people'))
-            'people'
-            >>> manager.model_for(manager.collection_name(Person))
-            <class 'mymodels.Person'>
-
-        """
-        # Reverse the dictionary.
-        models = {info.collection_name: model for model, info in self.created_apis_for.items()}
-        try:
-            return models[collection_name]
-        except KeyError:
-            raise ValueError('Collection name {0} unknown. Be sure to set the'
-                             ' `collection_name` keyword argument when calling'
-                             ' `create_api()`.'.format(collection_name))
+        self.include_links = include_links
 
     def url_for(self, model, **kw) -> str:
         """Returns the URL for the specified model, similar to
@@ -292,9 +182,12 @@ class APIManager(object):
         :class:`APIManager`.
 
         """
-        return self.created_apis_for[model].collection_name
+        try:
+            return self.created_apis_for[model].collection_name
+        except KeyError:
+            raise ValueError(f'Model is not registered with the API: {model}')
 
-    def serializer_for(self, model):
+    def serializer_for(self, model) -> Serializer:
         """Returns the serializer for the specified model, as specified
         by the `serializer` keyword argument to
         :meth:`create_api_blueprint`.
@@ -309,7 +202,7 @@ class APIManager(object):
         """
         return self.created_apis_for[model].serializer
 
-    def primary_key_for(self, model):
+    def primary_key_for(self, instance_or_model):
         """Returns the primary key for the specified model, as specified
         by the `primary_key` keyword argument to
         :meth:`create_api_blueprint`.
@@ -319,6 +212,12 @@ class APIManager(object):
         otherwise a :exc:`KeyError` is raised.
 
         """
+        # TODO: refactor to use model only
+        if isinstance(instance_or_model, type):
+            model = instance_or_model
+        else:
+            model = instance_or_model.__class__
+
         return self.created_apis_for[model].primary_key
 
     def url_prefix_for(self, model):
@@ -343,7 +242,7 @@ class APIManager(object):
         To use this method with pure SQLAlchemy, for example::
 
             from flask import Flask
-            from flask.ext.restless import APIManager
+            from flask_restless import APIManager
             from sqlalchemy import create_engine
             from sqlalchemy.orm.session import sessionmaker
 
@@ -355,20 +254,20 @@ class APIManager(object):
             ...
 
             # Create the API manager and create the APIs.
-            apimanager = APIManager(session=mysession)
-            apimanager.create_api(User)
-            apimanager.create_api(Comment)
+            api_manager = APIManager(session=mysession)
+            api_manager.create_api(User)
+            api_manager.create_api(Comment)
 
             # Later, call `init_app` to register the blueprints for the
             # APIs created earlier.
             app = Flask(__name__)
-            apimanager.init_app(app)
+            api_manager.init_app(app)
 
         and with models defined with Flask-SQLAlchemy::
 
             from flask import Flask
-            from flask.ext.restless import APIManager
-            from flask.ext.sqlalchemy import SQLAlchemy
+            from flask_restless import APIManager
+            from flask_sqlalchemy import SQLAlchemy
 
             db = SQLALchemy(app)
 
@@ -376,31 +275,44 @@ class APIManager(object):
             ...
 
             # Create the API manager and create the APIs.
-            apimanager = APIManager(flask_sqlalchemy_db=db)
-            apimanager.create_api(User)
-            apimanager.create_api(Comment)
+            api_manager = APIManager(session=db.session)
+            api_manager.create_api(User)
+            api_manager.create_api(Comment)
 
             # Later, call `init_app` to register the blueprints for the
             # APIs created earlier.
             app = Flask(__name__)
-            apimanager.init_app(app)
+            api_manager.init_app(app)
 
         """
         # Register any queued blueprints on the given application.
         for blueprint in self.blueprints:
             app.register_blueprint(blueprint)
 
-    def create_api_blueprint(self, name, model, methods=READONLY_METHODS,
-                             url_prefix=None, collection_name=None,
-                             allow_functions=False, only=None, exclude=None,
-                             additional_attributes=None,
-                             validation_exceptions=None, page_size=10,
-                             max_page_size=100, preprocessors=None,
-                             postprocessors=None, primary_key=None,
-                             serializer=None, deserializer=None,
-                             includes=None, allow_to_many_replacement=False,
-                             allow_delete_from_to_many_relationships=False,
-                             allow_client_generated_ids=False):
+    def create_api_blueprint(
+            self,
+            name: str,
+            model,
+            methods=READONLY_METHODS,
+            url_prefix: Optional[str] = None,
+            collection_name: Optional[str] = None,
+            only=None,
+            exclude=None,
+            additional_attributes=None,
+            validation_exceptions=None,
+            page_size: int = 10,
+            max_page_size: int = 100,
+            preprocessors=None,
+            postprocessors=None,
+            primary_key: Optional[str] = None,
+            serializer: Serializer = None,
+            deserializer: Deserializer = None,
+            includes=None,
+            allow_to_many_replacement: bool = False,
+            allow_delete_from_to_many_relationships: bool = False,
+            allow_client_generated_ids: bool = False,
+            allow_functions: bool = False
+    ):
         """Creates and returns a ReSTful API interface as a blueprint, but does
         not register it on any :class:`flask.Flask` application.
 
@@ -486,22 +398,11 @@ class APIManager(object):
         then this method creates endpoints of the form ``/api/foo``,
         ``/api/foo/<id>``, etc.
 
-        If `allow_functions` is ``True``, then :http:method:`get`
-        requests to ``/api/eval/<collection_name>`` will return the
-        result of evaluating SQL functions specified in the body of the
-        request. For information on the request format, see
-        :ref:`functionevaluation`. This is ``False`` by default.
-
-        .. warning::
-
-           If ``allow_functions`` is ``True``, you must not create an
-           API for a model whose name is ``'eval'``.
-
         If `only` is not ``None``, it must be a list of columns and/or
         relationships of the specified `model`, given either as strings or as
         the attributes themselves. If it is a list, only these fields will
         appear in the resource object representation of an instance of `model`.
-        In other words, `only` is a whitelist of fields. The ``id`` and
+        In other words, `only` is a allowlist of fields. The ``id`` and
         ``type`` elements of the resource object will always be present
         regardless of the value of this argument. If `only` contains a string
         that does not name a column in `model`, it will be ignored.
@@ -517,7 +418,7 @@ class APIManager(object):
         relationships of the specified `model`, given either as strings or as
         the attributes themselves. If it is a list, all fields **except** these
         will appear in the resource object representation of an instance of
-        `model`. In other words, `exclude` is a blacklist of fields. The ``id``
+        `model`. In other words, `exclude` is an blocklist of fields. The ``id``
         and ``type`` elements of the resource object will always be present
         regardless of the value of this argument. If `exclude` contains a
         string that does not name a column in `model`, it will be ignored.
@@ -545,15 +446,7 @@ class APIManager(object):
         see :ref:`pagination`.
 
         `serializer` and `deserializer` are custom serialization
-        functions. The former function must take a single positional
-        argument representing the instance of the model to serialize and
-        an additional keyword argument ``only`` representing the fields
-        to include in the serialized representation of the instance, and
-        must return a dictionary representation of that instance. The
-        latter function must take a single argument representing the
-        dictionary representation of an instance of the model and must
-        return an instance of `model` that has those attributes. For
-        more information, see :ref:`serialization`.
+        classes. See :ref:`serialization`.
 
         `preprocessors` is a dictionary mapping strings to lists of
         functions. Each key represents a type of endpoint (for example,
@@ -600,62 +493,76 @@ class APIManager(object):
         this be a UUID. This is ``False`` by default. For more information, see
         :ref:`creating`.
 
+        If `allow_functions` is ``True``, then :http:method:`get`
+        requests to ``/api/eval/<collection_name>`` will return the
+        result of evaluating SQL functions specified in the body of the
+        request. For information on the request format, see
+        :ref:`functionevaluation`. This is ``False`` by default.
+
+        .. warning::
+
+           This is deprecated and going to be removed in the next major version
+
+        .. warning::
+
+           If ``allow_functions`` is ``True``, you must not create an
+           API for a model whose name is ``'eval'``.
         """
         # Perform some sanity checks on the provided keyword arguments.
         if only is not None and exclude is not None:
             msg = 'Cannot simultaneously specify both `only` and `exclude`'
             raise IllegalArgumentError(msg)
-        if not hasattr(model, 'id') and len(primary_key_names(model)) == 0:
-            msg = 'Provided model must have an `id` attribute or primary key'
-            raise IllegalArgumentError(msg)
         if collection_name == '':
             msg = 'Collection name must be nonempty'
             raise IllegalArgumentError(msg)
         if collection_name is None:
             collection_name = model.__table__.name
+
         # convert all method names to upper case
         methods = frozenset((m.upper() for m in methods))
         # the name of the API, for use in creating the view and the blueprint
-        apiname = APIManager.api_name(collection_name)
+        api_name = f'{collection_name}_api'
         # Prepend the universal preprocessors and postprocessors specified in
         # the constructor of this class.
-        preprocessors_ = defaultdict(list)
-        postprocessors_ = defaultdict(list)
+        preprocessors_: Dict[str, list] = defaultdict(list)
+        postprocessors_: Dict[str, list] = defaultdict(list)
         preprocessors_.update(preprocessors or {})
         postprocessors_.update(postprocessors or {})
-        # for key, value in self.restless_info.universal_preprocessors.items():
+
         for key, value in self.pre.items():
             preprocessors_[key] = value + preprocessors_[key]
-        # for key, value in self.restless_info.universal_postprocessors.items():
+
         for key, value in self.post.items():
             postprocessors_[key] = value + postprocessors_[key]
+
         # Validate that all the additional attributes exist on the model.
         if additional_attributes is not None:
             for attr in additional_attributes:
-                if isinstance(attr, STRING_TYPES) and not hasattr(model, attr):
-                    msg = 'no attribute "{0}" on model {1}'.format(attr, model)
-                    raise AttributeError(msg)
+                if not hasattr(model, attr):
+                    raise AttributeError(f'no attribute "{attr}" on model {model}')
+
+        # find the primary_key of the model or try and use 'id'
+        if primary_key is None:
+            pk_names = primary_key_names(model)
+            primary_key = pk_names[0]
+
         # Create a default serializer and deserializer if none have been
         # provided.
         if serializer is None:
-            serializer = FastSerializer(model, collection_name, primary_key=primary_key,
-                                        only=only, exclude=exclude, additional_attributes=additional_attributes)
+            serializer = DefaultSerializer(model, collection_name, self, primary_key=primary_key,
+                                           only=only, exclude=exclude, additional_attributes=additional_attributes)
 
         session = self.session
         if deserializer is None:
-            deserializer = DefaultDeserializer(self.session, model,
-                                               allow_client_generated_ids)
+            deserializer = DefaultDeserializer(self.session, model, self, allow_client_generated_ids=allow_client_generated_ids)
         # Create the view function for the API for this model.
-        #
-        # Rename some variables with long names for the sake of brevity.
-        atmr = allow_to_many_replacement
-        api_view = API.as_view(apiname, session, model,
+        api_view = API.as_view(api_name, session, model, self,
                                # Keyword arguments for APIBase.__init__()
                                preprocessors=preprocessors_,
                                postprocessors=postprocessors_,
                                primary_key=primary_key,
                                validation_exceptions=validation_exceptions,
-                               allow_to_many_replacement=atmr,
+                               allow_to_many_replacement=allow_to_many_replacement,
                                # Keyword arguments for API.__init__()
                                page_size=page_size,
                                max_page_size=max_page_size,
@@ -667,25 +574,21 @@ class APIManager(object):
         # collection only, the second is for methods which may or may not
         # specify an instance, the third is for methods which must specify an
         # instance
-        # TODO what should the second argument here be?
-        # TODO should the url_prefix be specified here or in register_blueprint
-        if url_prefix is not None:
-            prefix = url_prefix
-        elif self.url_prefix is not None:
+
+        if url_prefix is None:
             prefix = self.url_prefix
         else:
-            prefix = DEFAULT_URL_PREFIX
+            prefix = url_prefix
+
         blueprint = Blueprint(name, __name__, url_prefix=prefix)
         add_rule = blueprint.add_url_rule
 
         # The URLs that will be routed below.
-        collection_url = '/{0}'.format(collection_name)
-        resource_url = '{0}/<resource_id>'.format(collection_url)
-        related_resource_url = '{0}/<relation_name>'.format(resource_url)
-        to_many_resource_url = \
-            '{0}/<related_resource_id>'.format(related_resource_url)
-        relationship_url = \
-            '{0}/relationships/<relation_name>'.format(resource_url)
+        collection_url = f'/{collection_name}'
+        resource_url = f'{collection_url}/<resource_id>'
+        related_resource_url = f'{resource_url}/<relation_name>'
+        to_many_resource_url = f'{related_resource_url}/<related_resource_id>'
+        relationship_url = f'{resource_url}/relationships/<relation_name>'
 
         # Create relationship URL endpoints.
         #
@@ -695,19 +598,19 @@ class APIManager(object):
         # :http:get:`/api/articles/1/relationships/author` interpret the
         # word `relationships` as the name of a relation of an article
         # object.
-        relationship_api_name = f'{apiname}_relationships'
-        rapi_view = RelationshipAPI.as_view
-        adftmr = allow_delete_from_to_many_relationships
-        relationship_api_view = \
-            rapi_view(relationship_api_name, session, model,
-                      # Keyword arguments for APIBase.__init__()
-                      preprocessors=preprocessors_,
-                      postprocessors=postprocessors_,
-                      primary_key=primary_key,
-                      validation_exceptions=validation_exceptions,
-                      allow_to_many_replacement=allow_to_many_replacement,
-                      # Keyword arguments RelationshipAPI.__init__()
-                      allow_delete_from_to_many_relationships=adftmr)
+        relationship_api_name = f'{api_name}_relationships'
+
+        relationship_api_view = RelationshipAPI.as_view(
+            relationship_api_name, session, model, self,
+            # Keyword arguments for APIBase.__init__()
+            preprocessors=preprocessors_,
+            postprocessors=postprocessors_,
+            primary_key=primary_key,
+            validation_exceptions=validation_exceptions,
+            allow_to_many_replacement=allow_to_many_replacement,
+            # Keyword arguments RelationshipAPI.__init__()
+            allow_delete_from_to_many_relationships=allow_delete_from_to_many_relationships
+        )
         # When PATCH is allowed, certain non-PATCH requests are allowed
         # on relationship URLs.
         relationship_methods = READONLY_METHODS & methods
@@ -716,6 +619,30 @@ class APIManager(object):
         add_rule(relationship_url, methods=relationship_methods,
                  view_func=relationship_api_view)
 
+        get_collection_function = FetchCollection.as_view(
+            name=f'{collection_name}_get_collection',
+            session=session,
+            model=model,
+            api_manager=self,
+            preprocessors=preprocessors_['GET_COLLECTION'],
+            postprocessors=postprocessors_['GET_COLLECTION'],
+            max_page_size=max_page_size,
+            page_size=page_size,
+            includes=includes
+        )
+        if 'GET' in methods:
+            add_rule(collection_url, view_func=get_collection_function, methods=['GET'])
+
+        get_resource_function = FetchResource.as_view(
+            name=f'{collection_name}_get_resource',
+            session=session,
+            model=model,
+            api_manager=self,
+            preprocessors=preprocessors_['GET_RESOURCE'],
+            postprocessors=postprocessors_['GET_RESOURCE'],
+            includes=includes
+        )
+
         # The URL for accessing the entire collection. (POST is special because
         # the :meth:`API.post` method doesn't have any arguments.)
         #
@@ -723,11 +650,6 @@ class APIManager(object):
         collection_methods = frozenset(('POST', )) & methods
         add_rule(collection_url, view_func=api_view,
                  methods=collection_methods)
-        collection_methods = frozenset(('GET', )) & methods
-        collection_defaults = dict(resource_id=None, relation_name=None,
-                                   related_resource_id=None)
-        add_rule(collection_url, view_func=api_view,
-                 methods=collection_methods, defaults=collection_defaults)
 
         # The URL for accessing a single resource. (DELETE and PATCH are
         # special because the :meth:`API.delete` and :meth:`API.patch` methods
@@ -737,9 +659,7 @@ class APIManager(object):
         resource_methods = frozenset(('DELETE', 'PATCH')) & methods
         add_rule(resource_url, view_func=api_view, methods=resource_methods)
         resource_methods = READONLY_METHODS & methods
-        resource_defaults = dict(relation_name=None, related_resource_id=None)
-        add_rule(resource_url, view_func=api_view, methods=resource_methods,
-                 defaults=resource_defaults)
+        add_rule(resource_url, view_func=get_resource_function, methods=resource_methods)
 
         # The URL for accessing a related resource, which may be a to-many or a
         # to-one relationship.
@@ -762,19 +682,42 @@ class APIManager(object):
         # which responds only to GET requests and responds with the result of
         # evaluating functions on all instances of the specified model
         if allow_functions:
-            eval_api_name = f'{apiname}_eval'
-            eval_api_view = FunctionAPI.as_view(eval_api_name, session, model)
-            eval_endpoint = '/eval{0}'.format(collection_url)
-            eval_methods = ['GET']
-            blueprint.add_url_rule(eval_endpoint, methods=eval_methods,
-                                   view_func=eval_api_view)
+            warnings.warn('`allow_functions` is deprecated and will be removed in the next major release', DeprecationWarning)
+            eval_api_view = FunctionAPI.as_view(f'{api_name}_eval', session, model)
+            blueprint.add_url_rule(f'/eval{collection_url}', methods=['GET'], view_func=eval_api_view)
 
         # Finally, record that this APIManager instance has created an API for
         # the specified model.
-        self.created_apis_for[model] = APIInfo(collection_name, blueprint.name,
-                                               serializer, primary_key, prefix)
+        api_info = registry.APIInfo(collection_name, blueprint.name, serializer, primary_key, prefix)
+        self.created_apis_for[model] = api_info
+        registry.add(model, api_info)
         return blueprint
 
+    def serialize_relationship(self, instance):
+        model = get_model(instance)
+        return {
+            'id': str(getattr(instance, self.primary_key_for(model))),
+            'type': self.collection_name(model)
+        }
+
+    def primary_key_value(self, instance, as_string=False):
+        """Returns the value of the primary key field of the specified `instance`
+        of a SQLAlchemy model.
+
+        This is a convenience function for::
+
+            getattr(instance, primary_key_name(instance))
+
+        If `as_string` is ``True``, try to coerce the return value to a string.
+
+        """
+
+        result = getattr(instance, self.primary_key_for(instance))
+        if not as_string:
+            return result
+
+        return str(result)
+
     def create_api(self, *args, **kw):
         """Creates and possibly registers a ReSTful API blueprint for
         the given SQLAlchemy model.
