@@ -20,16 +20,15 @@ relationship on a particular instance of a SQLAlchemy model.
 import inspect
 
 from sqlalchemy import and_
+from sqlalchemy import func
 from sqlalchemy import or_
 from sqlalchemy.orm import aliased
 from sqlalchemy.orm.attributes import InstrumentedAttribute
-from sqlalchemy.sql import false as FALSE
 
-from .helpers import get_model
+from .exceptions import BadRequest
 from .helpers import get_related_association_proxy_model
 from .helpers import get_related_model
 from .helpers import primary_key_names
-from .helpers import primary_key_value
 from .helpers import session_query
 from .helpers import string_to_datetime
 
@@ -37,7 +36,7 @@ try:
     # SQLAlchemy 1.3+
     from sqlalchemy.ext.associationproxy import ObjectAssociationProxyInstance as AssociationProxy
 except ImportError:
-    from sqlalchemy.ext.associationproxy import AssociationProxy
+    from sqlalchemy.ext.associationproxy import AssociationProxy  # type: ignore
 
 
 class ComparisonToNull(Exception):
@@ -137,6 +136,8 @@ OPERATORS = {
     'not_like': lambda f, a: ~f.like(a),
     'in': lambda f, a: f.in_(a),
     'not_in': lambda f, a: ~f.in_(a),
+    'to_tsquery': lambda f, a: f.match(a),
+    'plainto_tsquery': lambda f, a: f.op('@@')(func.plainto_tsquery(a)),
     # Operators which accept three arguments.
     'has': lambda f, a, fn: f.has(_sub_operator(f, a, fn)),
     'any': lambda f, a, fn: f.any(_sub_operator(f, a, fn)),
@@ -351,32 +352,11 @@ def create_filter(model, filt):
     return or_(create_filter(model, f) for f in filt)
 
 
-def search_relationship(session, instance, relation, filters=None, sort=None,
-                        group_by=None):
-    model = get_model(instance)
-    related_model = get_related_model(model, relation)
-    query = session_query(session, related_model)
-
-    # Filter by only those related values that are related to `instance`.
-    relationship = getattr(instance, relation)
-    primary_keys = {primary_key_value(inst) for inst in relationship}
-    # If the relationship is empty, we can avoid a potentially expensive
-    # filtering operation by simply returning an intentionally empty
-    # query.
-    if not primary_keys:
-        return query.filter(FALSE())
-    query = query.filter(primary_key_value(related_model).in_(primary_keys))
-
-    return search(session, related_model, filters=filters, sort=sort,
-                  group_by=group_by, _initial_query=query)
-
-
-def search(session, model, filters=None, sort=None, group_by=None,
-           _initial_query=None):
+def search(session, model, filters=None, sort=None, _initial_query=None):
     """Returns a SQLAlchemy query instance with the specified parameters.
 
     Each instance in the returned query meet the requirements specified by
-    ``filters``, ``sort``, and ``group_by``.
+    ``filters`` and ``sort``.
 
     This function returns a single instance of the model matching the search
     parameters if ``search_params['single']`` is ``True``, or a list of all
@@ -396,8 +376,7 @@ def search(session, model, filters=None, sort=None, group_by=None,
     will be appended to this query. Otherwise, an empty query will be
     created for the specified model.
 
-    When building the query, filters are applied first, then sorting, then
-    grouping.
+    When building the query, filters are applied first, then sorting.
 
     Raises :exc:`UnknownField` if one of the named fields given in one
     of the `filters` does not exist on the `model`.
@@ -412,12 +391,25 @@ def search(session, model, filters=None, sort=None, group_by=None,
     else:
         query = session_query(session, model)
 
-    # Filter the query.
-    filters = [Filter.from_dictionary(model, f) for f in filters]
-    # This function call may raise an exception.
-    filters = [create_filter(model, f) for f in filters]
+    try:
+        # Filter the query.
+        filters = [Filter.from_dictionary(model, f) for f in filters]
+
+        # This function call may raise an exception.
+        filters = [create_filter(model, f) for f in filters]
+    except UnknownField as e:
+        raise BadRequest(cause=e, details=f'Invalid filter object: No such field "{e.field}"') from e
+    except Exception as e:
+        raise BadRequest(cause=e, details='Unable to construct query') from e
+
     query = query.filter(*filters)
 
+    def get_field(obj, name):
+        try:
+            return getattr(obj, name)
+        except AttributeError:
+            raise BadRequest(details=f'Invalid sorting: No such field {name}')
+
     # Order the query. If no order field is specified, order by primary
     # key.
     # if not _ignore_sort:
@@ -427,12 +419,12 @@ def search(session, model, filters=None, sort=None, group_by=None,
             if '.' in field_name:
                 field_name, field_name_in_relation = field_name.split('.')
                 relation_model = aliased(get_related_model(model, field_name))
-                field = getattr(relation_model, field_name_in_relation)
+                field = get_field(relation_model, field_name_in_relation)
                 direction = getattr(field, direction_name)
                 query = query.join(relation_model)
                 query = query.order_by(direction())
             else:
-                field = getattr(model, field_name)
+                field = get_field(model, field_name)
                 direction = getattr(field, direction_name)
                 query = query.order_by(direction())
     else:
@@ -440,17 +432,4 @@ def search(session, model, filters=None, sort=None, group_by=None,
         pk_order = (getattr(model, field).asc() for field in pks)
         query = query.order_by(*pk_order)
 
-    # Group the query.
-    if group_by:
-        for field_name in group_by:
-            if '.' in field_name:
-                field_name, field_name_in_relation = field_name.split('.')
-                relation_model = get_related_model(model, field_name)
-                field = getattr(relation_model, field_name_in_relation)
-                query = query.join(relation_model)
-                query = query.group_by(field)
-            else:
-                field = getattr(model, field_name)
-                query = query.group_by(field)
-
     return query
