@@ -19,42 +19,41 @@ deserialization as expected by classes that follow the JSON API
 protocol.
 
 """
-from __future__ import division
-
+import enum
+from abc import ABC
+from abc import abstractmethod
+from collections import namedtuple
 from copy import copy
 from datetime import date
 from datetime import datetime
 from datetime import time
 from datetime import timedelta
+from typing import Any
+from typing import Dict
+from typing import FrozenSet
+from typing import Optional
 from urllib.parse import urljoin
 
 from flask import request
-from sqlalchemy import Column
-from sqlalchemy.exc import NoInspectionAvailable
-from sqlalchemy.ext.hybrid import HYBRID_PROPERTY
-from sqlalchemy.inspection import inspect
-from werkzeug.urls import url_quote_plus
+from sqlalchemy import inspect
+from sqlalchemy.orm.base import MANYTOONE
 
 from .helpers import attribute_columns
-from .helpers import collection_name
 from .helpers import foreign_keys
 from .helpers import get_by
-from .helpers import get_model
+from .helpers import get_column_name
 from .helpers import get_related_model
 from .helpers import get_relations
 from .helpers import has_field
 from .helpers import is_like_list
-from .helpers import is_mapped_class
-from .helpers import primary_key_for
 from .helpers import primary_key_names
-from .helpers import primary_key_value
-from .helpers import serializer_for
 from .helpers import strings_to_datetimes
-from .helpers import url_for
 
 #: Names of columns which should definitely not be considered user columns to
 #: be included in a dictionary representation of a model.
-COLUMN_BLACKLIST = ('_sa_polymorphic_on', )
+COLUMN_EXCLUDE_LIST = {'_sa_polymorphic_on'}
+
+RelationshipInfo = namedtuple('RelationshipInfo', ['foreign_key', 'target_model'])
 
 
 class SerializationException(Exception):
@@ -75,11 +74,13 @@ class SerializationException(Exception):
 
     """
 
-    def __init__(self, instance, message=None, resource=None, *args, **kw):
+    def __init__(self, instance, message=None, resource=None, resource_type=None, resource_id=None, *args, **kw):
         super(SerializationException, self).__init__(*args, **kw)
         self.resource = resource
         self.message = message
         self.instance = instance
+        self.resource_type = resource_type
+        self.resource_id = resource_id
 
 
 class DeserializationException(Exception):
@@ -174,7 +175,7 @@ class UnknownField(DeserializationException):
     #:
     #: This attribute can only take one of the two values ``'field'`` or
     #: ``'relationship'``.
-    field_type = None
+    field_type: Optional[str] = None
 
     def __init__(self, field, *args, **kw):
         super(UnknownField, self).__init__(*args, **kw)
@@ -213,7 +214,7 @@ class MissingInformation(DeserializationException):
     #: The name of the key in the dictionary that is missing.
     #:
     #: Subclasses must set this class attribute.
-    element = None
+    element: Optional[str] = None
 
     def __init__(self, relation_name=None, *args, **kw):
         super(MissingInformation, self).__init__(*args, **kw)
@@ -255,91 +256,24 @@ class MissingType(MissingInformation):
     element = 'type'
 
 
-def get_column_name(column):
-    """Retrieve a column name from a column attribute of SQLAlchemy model
-    class, or a string.
-
-    Raises `TypeError` when argument does not fall into either of those
-    options.
-
-    """
-    # TODO use inspection API here
-    if hasattr(column, '__clause_element__'):
-        clause_element = column.__clause_element__()
-        if not isinstance(clause_element, Column):
-            msg = 'Expected a column attribute of a SQLAlchemy ORM class'
-            raise TypeError(msg)
-        return clause_element.key
-    return column
-
-
-def create_relationship(model, instance, relation):
-    """Creates a relationship from the given relation name.
-
-    Returns a dictionary representing a relationship as described in
-    the `Relationships`_ section of the JSON API specification.
-
-    `model` is the model class of the primary resource for which a
-    relationship object is being created.
-
-    `instance` is the instance of the model for which we are considering
-    a related value.
-
-    `relation` is the name of the relation of `instance` given as a
-    string.
-
-    This function may raise :exc:`ValueError` if an API has not been
-    created for the primary model, `model`, or the model of the
-    relation.
-
-    .. _Relationships: http://jsonapi.org/format/#document-resource-object-relationships
-
-    """
-    result = {}
-    # Create the self and related links.
-    pk_value = primary_key_value(instance)
-    self_link = url_for(model, pk_value, relation, relationship=True)
-    related_link = url_for(model, pk_value, relation)
-    result['links'] = {'self': self_link}
-    # If the user has not created a GET endpoint for the related
-    # resource, then there is no "related" link to provide, so we check
-    # whether the URL exists before setting the related link.
-    try:
-        get_related_model(model, relation)
-    except ValueError:
-        pass
-    else:
-        result['links']['related'] = related_link
-    # Get the related value so we can see if it is a to-many
-    # relationship or a to-one relationship.
-    related_value = getattr(instance, relation)
-    # There are three possibilities for the relation: it could be a
-    # to-many relationship, a null to-one relationship, or a non-null
-    # to-one relationship. We decide whether the relation is to-many by
-    # determining whether it is list-like.
-    if is_like_list(instance, relation):
-        # We could pre-compute the "type" name for the related instances
-        # here and provide it in the `_type` keyword argument to the
-        # serialization function, but the to-many relationship could be
-        # heterogeneous.
-        result['data'] = [simple_relationship_serialize(instance)
-                          for instance in related_value]
-    elif related_value is not None:
-        result['data'] = simple_relationship_serialize(related_value)
-    else:
-        result['data'] = None
-    return result
-
-
-class Serializer(object):
+class Serializer(ABC):
     """An object that, when called, returns a dictionary representation of a
     given instance of a SQLAlchemy model.
 
-    **This is a base class with no implementation.**
-
     """
 
-    def __call__(self, instance, only=None):
+    @property
+    @abstractmethod
+    def attributes_columns(self) -> FrozenSet[str]:
+        """Provides a set of column names that should represent model's attributes"""
+
+    @property
+    @abstractmethod
+    def relationship_columns(self) -> FrozenSet[str]:
+        """Provides a set of column names that should represent model's relationships"""
+
+    @abstractmethod
+    def serialize(self, instance, only=None):
         """Returns a dictionary representation of the specified instance of a
         SQLAlchemy model.
 
@@ -347,15 +281,10 @@ class Serializer(object):
         appear as strings in `only` should appear in the returned
         dictionary. The only exception is that the keys ``'id'`` and ``'type'``
         will always appear, regardless of whether they appear in `only`.
-
-        **This method is not implemented in this base class; subclasses must
-        override this method.**
-
         """
-        raise NotImplementedError
 
 
-class Deserializer(object):
+class Deserializer(ABC):
     """An object that, when called, returns an instance of the SQLAlchemy model
     specified at instantiation time.
 
@@ -369,11 +298,13 @@ class Deserializer(object):
 
     """
 
-    def __init__(self, session, model):
+    def __init__(self, session, model, api_manager):
         self.session = session
         self.model = model
+        self.api_manager = api_manager
 
-    def __call__(self, document):
+    @abstractmethod
+    def deserialize(self, document):
         """Creates and returns a new instance of the SQLAlchemy model specified
         in the constructor whose attributes are given by the specified
         dictionary.
@@ -384,239 +315,15 @@ class Deserializer(object):
         see the `Resource Objects`_ section of the JSON API
         specification.
 
-        **This method is not implemented in this base class; subclasses must
-        override this method.**
-
         .. _Resource Objects: http://jsonapi.org/format/#document-structure-resource-objects
 
         """
-        raise NotImplementedError
 
 
 class DefaultSerializer(Serializer):
-    """A default implementation of a serializer for SQLAlchemy models.
+    """Default Serializer implementation."""
 
-    When called, this object returns a dictionary representation of a given
-    SQLAlchemy instance that meets the requirements of the JSON API
-    specification.
-
-    If `only` is a list, only these fields and relationships will in the
-    returned dictionary. The only exception is that the keys ``'id'`` and
-    ``'type'`` will always appear, regardless of whether they appear in `only`.
-    These settings take higher priority than the `only` list provided to the
-    :meth:`__call__` method: if an attribute or relationship appears in the
-    `only` argument to :meth:`__call__` but not here in the constructor, it
-    will not appear in the returned dictionary.
-
-    If `exclude` is a list, these fields and relationships will **not** appear
-    in the returned dictionary.
-
-    If `additional_attributes` is a list, these attributes of the instance to
-    be serialized will appear in the returned dictionary. This is useful if
-    your model has an attribute that is not a SQLAlchemy column but you want it
-    to be exposed.
-
-    If both `only` and `exclude` are specified, a :exc:`ValueError` is raised.
-    Also, if any attributes specified in `additional_attributes` appears in
-    `exclude`, a :exc:`ValueError` is raised.
-
-    """
-
-    def __init__(self, only=None, exclude=None, additional_attributes=None,
-                 **kw):
-        super(DefaultSerializer, self).__init__(**kw)
-        if only is not None and exclude is not None:
-            raise ValueError('Cannot specify both `only` and `exclude` keyword'
-                             ' arguments simultaneously')
-        if (additional_attributes is not None and exclude is not None and
-                any(attr in exclude for attr in additional_attributes)):
-            raise ValueError('Cannot exclude attributes listed in the'
-                             ' `additional_attributes` keyword argument')
-        # Always include at least the type and ID, regardless of what the user
-        # specified.
-        if only is not None:
-            # Convert SQLAlchemy Column objects to strings if necessary.
-            only = {get_column_name(column) for column in only}
-            only |= {'type', 'id'}
-        if exclude is not None:
-            # Convert SQLAlchemy Column objects to strings if necessary.
-            exclude = {get_column_name(column) for column in exclude}
-        self.default_fields = only
-        self.exclude = exclude
-        self.additional_attributes = additional_attributes
-
-    def __call__(self, instance, only=None):
-        """Returns a dictionary representing the fields of the specified
-        instance of a SQLAlchemy model.
-
-        The returned dictionary is suitable as an argument to
-        :func:`flask.jsonify`; datetime objects (:class:`datetime.date`,
-        :class:`datetime.time`, :class:`datetime.datetime`, and
-        :class:`datetime.timedelta`) as well as :class:`uuid.UUID`
-        objects are converted to string representations, so no special
-        JSON encoder behavior is required.
-
-        If `only` is a list, only the fields and relationships whose
-        names appear as strings in `only` will appear in the resulting
-        dictionary. This filter is applied *after* the default fields
-        specified in the `only` keyword argument to the constructor of
-        this class, so only fields that appear in both `only` keyword
-        arguments will appear in the returned dictionary. The only
-        exception is that the keys ``'id'`` and ``'type'`` will always
-        appear, regardless of whether they appear in `only`.
-
-        Since this function creates absolute URLs to resources linked to the
-        given instance, it must be called within a `Flask request context`_.
-
-        .. _Flask request context: http://flask.pocoo.org/docs/0.10/reqcontext/
-
-        """
-        # Always include at least the type, ID, and the self link, regardless
-        # of what the user requested.
-        if only is not None:
-            only = set(only) | {'type', 'id'}
-        model = type(instance)
-        try:
-            inspected_instance = inspect(model)
-        except NoInspectionAvailable:
-            return instance
-        column_attrs = inspected_instance.column_attrs.keys()
-        descriptors = inspected_instance.all_orm_descriptors.items()
-        # hybrid_columns = [k for k, d in descriptors
-        #                   if d.extension_type == hybrid.HYBRID_PROPERTY
-        #                   and not (deep and k in deep)]
-        hybrid_columns = [k for k, d in descriptors
-                          if d.extension_type == HYBRID_PROPERTY]
-        columns = column_attrs + hybrid_columns
-        # Also include any attributes specified by the user.
-        if self.additional_attributes is not None:
-            columns += self.additional_attributes
-
-        # Only include fields allowed by the user during the instantiation of
-        # this object.
-        if self.default_fields is not None:
-            columns = (c for c in columns if c in self.default_fields)
-        # If `only` is a list, only include those columns that are in the list.
-        if only is not None:
-            columns = (c for c in columns if c in only)
-
-        # Exclude columns specified by the user during the instantiation of
-        # this object.
-        if self.exclude is not None:
-            columns = (c for c in columns if c not in self.exclude)
-        # Exclude column names that are blacklisted.
-        columns = (c for c in columns
-                   if not c.startswith('__') and c not in COLUMN_BLACKLIST)
-        # Exclude column names that are foreign keys.
-        foreign_key_columns = foreign_keys(model)
-        columns = (c for c in columns if c not in foreign_key_columns)
-
-        # Create a dictionary mapping attribute name to attribute value for
-        # this particular instance.
-        attributes = {column: getattr(instance, column) for column in columns}
-        # Call any functions that appear in the result.
-        #
-        attributes = {k: (v() if callable(v) else v) for k, v in attributes.items()}
-        # Serialize any date- or time-like objects that appear in the
-        # attributes.
-        #
-        # TODO In Flask 1.0, the default JSON encoder for the Flask
-        # application object does this automatically. Alternately, the
-        # user could have set a smart JSON encoder on the Flask
-        # application, which would cause these attributes to be
-        # converted to strings when the Response object is created (in
-        # the `jsonify` function, for example). However, we should not
-        # rely on that JSON encoder since the user could set any crazy
-        # encoder on the Flask application.
-        for key, val in attributes.items():
-            if isinstance(val, (date, datetime, time)):
-                attributes[key] = val.isoformat()
-            elif isinstance(val, timedelta):
-                attributes[key] = val.total_seconds()
-        # Recursively serialize any object that appears in the
-        # attributes. This may happen if, for example, the return value
-        # of one of the callable functions is an instance of another
-        # SQLAlchemy model class.
-        for key, val in attributes.items():
-            # This is a bit of a fragile test for whether the object
-            # needs to be serialized: we simply check if the class of
-            # the object is a mapped class.
-            if is_mapped_class(type(val)):
-                model_ = get_model(val)
-                try:
-                    serialize = serializer_for(model_)
-                except ValueError:
-                    # TODO Should this cause an exception, or fail
-                    # silently? See similar comments in `views/base.py`.
-                    # # raise SerializationException(instance)
-                    serialize = simple_serialize
-                attributes[key] = serialize(val)
-        # Get the ID and type of the resource.
-        id_ = attributes.pop('id')
-        type_ = collection_name(model)
-        # Create the result dictionary and add the attributes.
-        result = dict(id=id_, type=type_)
-        if attributes:
-            result['attributes'] = attributes
-        # Add the self link unless it has been explicitly excluded.
-        if ((self.default_fields is None or 'self' in self.default_fields)
-                and (only is None or 'self' in only)):
-            instance_id = primary_key_value(instance)
-            path = url_for(model, instance_id, _method='GET')
-            url = urljoin(request.url_root, path)
-            result['links'] = dict(self=url)
-        # # add any included methods
-        # if include_methods is not None:
-        #     for method in include_methods:
-        #         if '.' not in method:
-        #             value = getattr(instance, method)
-        #             # Allow properties and static attributes in
-        #             # include_methods
-        #             if callable(value):
-        #                 value = value()
-        #             result[method] = value
-
-        # If the primary key is not named "id", we'll duplicate the
-        # primary key under the "id" key.
-        pk_name = primary_key_for(model)
-        if pk_name != 'id':
-            result['id'] = result['attributes'][pk_name]
-        # TODO Same problem as above.
-        #
-        # In order to comply with the JSON API standard, primary keys must be
-        # returned to the client as strings, so we convert it here.
-        if 'id' in result:
-            try:
-                result['id'] = str(result['id'])
-            except UnicodeEncodeError:
-                result['id'] = url_quote_plus(result['id'].encode('utf-8'))
-        # If there are relations to convert to dictionary form, put them into a
-        # special `links` key as required by JSON API.
-        relations = get_relations(model)
-        if self.default_fields is not None:
-            relations = [r for r in relations if r in self.default_fields]
-        # Only consider those relations listed in `only`.
-        if only is not None:
-            relations = [r for r in relations if r in only]
-        # Exclude relations specified by the user during the instantiation of
-        # this object.
-        if self.exclude is not None:
-            relations = [r for r in relations if r not in self.exclude]
-        if not relations:
-            return result
-        # For the sake of brevity, rename this function.
-        cr = create_relationship
-        result['relationships'] = {rel: cr(model, instance, rel) for rel in relations}
-        return result
-
-
-class FastSerializer(Serializer):
-    """An implementation of serializer optimized for speed.
-
-    Instead of inspecting each model (attributes, foreign keys, etc.) during serialization, it does that during instantiation.
-    """
-
-    def __init__(self, model, type_name, primary_key=None, only=None, exclude=None, additional_attributes=None, **kwargs):
+    def __init__(self, model, type_name, api_manager, primary_key=None, only=None, exclude=None, additional_attributes=None, **kwargs):
         super().__init__(**kwargs)
         if only is not None and exclude is not None:
             raise ValueError('Cannot specify both `only` and `exclude` keyword arguments simultaneously')
@@ -624,11 +331,16 @@ class FastSerializer(Serializer):
         if additional_attributes is not None and exclude is not None and any(attr in exclude for attr in additional_attributes):
             raise ValueError('Cannot exclude attributes listed in the `additional_attributes` keyword argument')
 
+        self._api_manager = api_manager
         self._model = model
         self._type = type_name
-        if primary_key is None:
-            pk_names = primary_key_names(model)
-            primary_key = 'id' if 'id' in pk_names else pk_names[0]
+        pk_names = primary_key_names(model)
+        if primary_key:
+            if primary_key not in pk_names:
+                raise ValueError(f'Column `{primary_key}` is not a primary key')
+        else:
+            raise ValueError('`primary_key` is required')
+
         self._primary_key = primary_key
         self._relations = set(get_relations(model))
         self._only = None
@@ -662,14 +374,36 @@ class FastSerializer(Serializer):
         # https://jsonapi.org/format/#document-resource-object-attributes
         columns -= set(foreign_keys(model))
 
-        # Exclude column names that are blacklisted.
-        columns = {column for column in columns if not column.startswith('__') and column not in COLUMN_BLACKLIST}
+        # Exclude column names that are on the exclude list.
+        columns = {column for column in columns if not column.startswith('__') and column not in COLUMN_EXCLUDE_LIST}
 
-        self._columns = columns
+        self._relations = frozenset(self._relations)
+        self._columns = frozenset(columns)
 
-    def __call__(self, instance, only=None):
-        columns = copy(self._columns)
+        # Finding ManyToOne relationships that can be rendered using FK
+        inspected_model = inspect(model)
+        self._many_to_one_relationships = {}
 
+        for relationship in inspected_model.relationships:
+            if relationship.key in self._relations and relationship.direction == MANYTOONE:
+                foreign_key = relationship.local_remote_pairs[0][0].key
+                target_model = relationship.mapper.class_
+                self._many_to_one_relationships[relationship.key] = RelationshipInfo(foreign_key=foreign_key, target_model=target_model)
+
+    @property
+    def many_to_one_relationships(self):
+        return set(self._many_to_one_relationships.keys())
+
+    @property
+    def relationship_columns(self):
+        return self._relations
+
+    @property
+    def attributes_columns(self):
+        return self._columns
+
+    def serialize_attributes(self, instance, only=None) -> Dict[str, Any]:
+        columns = self.attributes_columns
         if only is not None:
             columns &= only
 
@@ -678,46 +412,24 @@ class FastSerializer(Serializer):
         attributes = {column: getattr(instance, column) for column in columns}
 
         for key, value in attributes.items():
-            # Call any functions that appear in the result.
-            if callable(value):
-                attributes[key] = value()
-
-        # Serialize any date- or time-like objects that appear in the
-        # attributes.
-        #
-        # TODO In Flask 1.0, the default JSON encoder for the Flask
-        # application object does this automatically. Alternately, the
-        # user could have set a smart JSON encoder on the Flask
-        # application, which would cause these attributes to be
-        # converted to strings when the Response object is created (in
-        # the `jsonify` function, for example). However, we should not
-        # rely on that JSON encoder since the user could set any crazy
-        # encoder on the Flask application.
-        for key, value in attributes.items():
+            # Serialize any date- or time-like objects that appear in the attributes.
             if isinstance(value, (date, datetime, time)):
                 attributes[key] = value.isoformat()
             elif isinstance(value, timedelta):
                 attributes[key] = value.total_seconds()
+            # Enums are not serializable by default, use 'name' property
+            elif isinstance(value, enum.Enum):
+                attributes[key] = value.name
 
-        # TODO: Drop this in the future releases
-        # Recursively serialize any object that appears in the
-        # attributes. This may happen if, for example, the return value
-        # of one of the callable functions is an instance of another
-        # SQLAlchemy model class.
-        for key, val in attributes.items():
-            # This is a bit of a fragile test for whether the object
-            # needs to be serialized: we simply check if the class of
-            # the object is a mapped class.
-            if is_mapped_class(type(val)):
-                model_ = get_model(val)
-                try:
-                    serialize = serializer_for(model_)
-                except ValueError:
-                    # TODO Should this cause an exception, or fail
-                    # silently? See similar comments in `views/base.py`.
-                    # # raise SerializationException(instance)
-                    serialize = simple_serialize
-                attributes[key] = serialize(val)
+        return attributes
+
+    def serialize(self, instance, only=None):
+        attributes = self.serialize_attributes(instance, only)
+
+        for key, value in attributes.items():
+            # Call any functions that appear in the result.
+            if callable(value):
+                attributes[key] = value()
 
         # Get the ID and type of the resource.
         id_ = str(getattr(instance, self._primary_key))
@@ -733,33 +445,85 @@ class FastSerializer(Serializer):
             relations &= only
 
         if relations:
-            result['relationships'] = {rel: create_relationship(self._model, instance, rel) for rel in relations}
+            result['relationships'] = {rel: self.create_relationship(instance, rel) for rel in relations}
 
         # TODO: Refactor
-        if ((self._only is None or 'self' in self._only)
+        if (self._api_manager.include_links and (self._only is None or 'self' in self._only)
                 and (only is None or 'self' in only)):
-            instance_id = primary_key_value(instance)
-            path = url_for(self._model, instance_id, _method='GET')
+            instance_id = getattr(instance, self._api_manager.primary_key_for(self._model))
+            path = self._api_manager.url_for(self._model, resource_id=instance_id, _method='GET')
             url = urljoin(request.url_root, path)
             result['links'] = dict(self=url)
 
         return result
 
+    def create_relationship(self, instance, relation):
+        """Creates a relationship from the given relation name.
 
-class DefaultRelationshipSerializer(Serializer):
-    """A default implementation of a serializer for resource identifier
-    objects for use in relationship objects in JSON API documents.
+        Returns a dictionary representing a relationship as described in
+        the `Relationships`_ section of the JSON API specification.
 
-    This serializer differs from the default serializer for resources
-    since it only provides an ``'id'`` and a ``'type'`` in the
-    dictionary returned by the :meth:`__call__` method.
+        `instance` is the instance of the model for which we are considering
+        a related value.
 
-    """
+        `relation` is the name of the relation of `instance` given as a
+        string.
 
-    def __call__(self, instance, only=None, _type=None):
-        if _type is None:
-            _type = collection_name(get_model(instance))
-        return {'id': str(primary_key_value(instance)), 'type': _type}
+        This function may raise :exc:`ValueError` if an API has not been
+        created for the primary model, `model`, or the model of the
+        relation.
+
+        .. _Relationships: http://jsonapi.org/format/#document-resource-object-relationships
+
+        """
+        result = {}
+        if self._api_manager.include_links:
+            # Create the self and related links.
+            pk_value = getattr(instance, self._api_manager.primary_key_for(self._model))
+            self_link = self._api_manager.url_for(self._model, resource_id=pk_value, relation_name=relation, relationship=True)
+            related_link = self._api_manager.url_for(self._model, resource_id=pk_value, relation_name=relation)
+            result['links'] = {'self': self_link}
+            # If the user has not created a GET endpoint for the related
+            # resource, then there is no "related" link to provide, so we check
+            # whether the URL exists before setting the related link.
+            try:
+                get_related_model(self._model, relation)
+            except ValueError:
+                pass
+            else:
+                result['links']['related'] = related_link
+
+        relationship_info = self._many_to_one_relationships.get(relation)
+        if relationship_info:
+            fk_value = getattr(instance, relationship_info.foreign_key)
+            if fk_value is None:
+                result['data'] = None
+            else:
+                result['data'] = {
+                    'id': str(fk_value),
+                    'type': self._api_manager.collection_name(relationship_info.target_model)
+                }
+            return result
+
+        # Get the related value so we can see if it is a to-many
+        # relationship or a to-one relationship.
+        related_value = getattr(instance, relation)
+        # There are three possibilities for the relation: it could be a
+        # to-many relationship, a null to-one relationship, or a non-null
+        # to-one relationship. We decide whether the relation is to-many by
+        # determining whether it is list-like.
+        if is_like_list(instance, relation):
+            # We could pre-compute the "type" name for the related instances
+            # here and provide it in the `_type` keyword argument to the
+            # serialization function, but the to-many relationship could be
+            # heterogeneous.
+            result['data'] = [self._api_manager.serialize_relationship(instance)
+                              for instance in related_value]
+        elif related_value is not None:
+            result['data'] = self._api_manager.serialize_relationship(related_value)
+        else:
+            result['data'] = None
+        return result
 
 
 class DefaultDeserializer(Deserializer):
@@ -770,13 +534,13 @@ class DefaultDeserializer(Deserializer):
 
     """
 
-    def __init__(self, session, model, allow_client_generated_ids=False, **kw):
-        super(DefaultDeserializer, self).__init__(session, model, **kw)
+    def __init__(self, *args, allow_client_generated_ids=False, **kw):
+        super(DefaultDeserializer, self).__init__(*args, **kw)
 
         #: Whether to allow client generated IDs.
         self.allow_client_generated_ids = allow_client_generated_ids
 
-    def __call__(self, document):
+    def deserialize(self, document):
         """Creates and returns an instance of the SQLAlchemy model
         specified in the constructor.
 
@@ -795,7 +559,7 @@ class DefaultDeserializer(Deserializer):
         if 'id' in data and not self.allow_client_generated_ids:
             raise ClientGeneratedIDNotAllowed
         type_ = data.pop('type')
-        expected_type = collection_name(self.model)
+        expected_type = self.api_manager.collection_name(self.model)
         if type_ != expected_type:
             raise ConflictingType(expected_type, type_)
         # Check for any request parameter naming a column which does not exist
@@ -818,11 +582,10 @@ class DefaultDeserializer(Deserializer):
                     raise MissingData(link_name)
                 linkage = link_object['data']
                 related_model = get_related_model(self.model, link_name)
-                expected_type = collection_name(related_model)
+                expected_type = self.api_manager.collection_name(related_model)
                 # Create the deserializer for this relationship object.
-                DRD = DefaultRelationshipDeserializer
-                deserialize = DRD(self.session, related_model, link_name)
-                links[link_name] = deserialize(linkage)
+                deserializer = DefaultRelationshipDeserializer(self.session, related_model, self.api_manager, relation_name=link_name)
+                links[link_name] = deserializer.deserialize(linkage)
         # TODO Need to check here if any related instances are None,
         # like we do in the patch() method. We could possibly refactor
         # the code above and the code there into a helper function...
@@ -866,19 +629,16 @@ class DefaultRelationshipDeserializer(Deserializer):
 
     """
 
-    def __init__(self, session, model, relation_name=None):
-        super(DefaultRelationshipDeserializer, self).__init__(session, model)
-        #: The related model whose objects this deserializer will return
-        #: in the :meth:`__call__` method.
-        self.model = model
+    def __init__(self, *args, relation_name=None):
+        super(DefaultRelationshipDeserializer, self).__init__(*args)
 
         #: The collection name given to the related model.
-        self.type_name = collection_name(self.model)
+        self.type_name = self.api_manager.collection_name(self.model)
 
         #: The name of the relationship being deserialized, as a string.
         self.relation_name = relation_name
 
-    def __call__(self, data):
+    def deserialize(self, data):
         """Gets the resource associated with the given resource
         identifier object.
 
@@ -906,22 +666,8 @@ class DefaultRelationshipDeserializer(Deserializer):
                 raise ConflictingType(self.relation_name, self.type_name,
                                       type_)
             id_ = data['id']
-            return get_by(self.session, self.model, id_)
+            primary_key = self.api_manager.primary_key_for(self.model)
+            return get_by(self.session, self.model, id_, primary_key=primary_key)
         # Otherwise, if this is a to-many relationship, recurse on each
         # and return a list of instances.
-        return list(map(self, data))
-
-
-#: Provides basic, uncustomized serialization functionality as provided by
-#: :class:`DefaultSerializer`.
-#:
-#: This function is suitable for calling on its own, no other instantiation or
-#: customization necessary.
-simple_serialize = DefaultSerializer()
-
-
-#: Basic serializer for relationship objects.
-#:
-#: This function is suitable for calling on its own, no other instantiation or
-#: customization necessary.
-simple_relationship_serialize = DefaultRelationshipSerializer()
+        return list(map(self.deserialize, data))
