@@ -16,16 +16,16 @@ The main class in this module, :class:`RelationshipAPI`, is a
 relationships according to the JSON API specification.
 
 """
+from flask import escape
 from flask import json
 from flask import request
 from werkzeug.exceptions import BadRequest
 
-from ..helpers import collection_name
 from ..helpers import get_by
 from ..helpers import get_related_model
 from ..helpers import is_like_list
 from .base import APIBase
-from .base import SingleKeyError
+from .base import collection_parameters
 from .base import error
 from .base import error_response
 from .base import errors_response
@@ -48,9 +48,9 @@ class RelationshipAPI(APIBase):
 
     """
 
-    def __init__(self, session, model,
+    def __init__(self, session, model, api_manager,
                  allow_delete_from_to_many_relationships=False, *args, **kw):
-        super(RelationshipAPI, self).__init__(session, model, *args, **kw)
+        super(RelationshipAPI, self).__init__(session, model, api_manager, *args, **kw)
         self.allow_delete_from_to_many_relationships = \
             allow_delete_from_to_many_relationships
 
@@ -90,22 +90,12 @@ class RelationshipAPI(APIBase):
         primary_resource = get_by(self.session, self.model, resource_id,
                                   self.primary_key)
         if primary_resource is None:
-            detail = 'No resource with ID {0}'.format(resource_id)
-            return error_response(404, detail=detail)
+            return error_response(404, detail=f'No resource with ID {escape(resource_id)}')
         if is_like_list(primary_resource, relation_name):
-            try:
-                filters, sort, group_by, single = self._collection_parameters()
-            except (TypeError, ValueError, OverflowError) as exception:
-                detail = 'Unable to decode filter objects as JSON list'
-                return error_response(400, cause=exception, detail=detail)
-            except SingleKeyError as exception:
-                detail = 'Invalid format for filter[single] query parameter'
-                return error_response(400, cause=exception, detail=detail)
+            filters, sort = collection_parameters()
             return self._get_collection_helper(resource=primary_resource,
                                                relation_name=relation_name,
-                                               filters=filters, sort=sort,
-                                               group_by=group_by,
-                                               single=single)
+                                               filters=filters, sort=sort)
         resource = getattr(primary_resource, relation_name)
         return self._get_resource_helper(resource,
                                          primary_resource=primary_resource,
@@ -131,43 +121,33 @@ class RelationshipAPI(APIBase):
             # See the note under the preprocessor in the get() method.
             if temp_result is not None:
                 resource_id, relation_name = temp_result
-        instance = get_by(self.session, self.model, resource_id,
-                          self.primary_key)
+        instance = get_by(self.session, self.model, resource_id, self.primary_key)
         # If no instance of the model exists with the specified instance ID,
         # return a 404 response.
         if instance is None:
-            detail = 'No instance with ID {0} in model {1}'
-            detail = detail.format(resource_id, self.model)
-            return error_response(404, detail=detail)
+            return error_response(404, detail=f'No instance with ID {escape(resource_id)} in model {self.model}')
         # If no such relation exists, return a 404.
         if not hasattr(instance, relation_name):
-            detail = 'Model {0} has no relation named {1}'
-            detail = detail.format(self.model, relation_name)
-            return error_response(404, detail=detail)
+            return error_response(404, detail=f'Model {self.model} has no relation named {escape(relation_name)}')
         related_model = get_related_model(self.model, relation_name)
         related_value = getattr(instance, relation_name)
         # Unwrap the data from the request.
         data = data.pop('data', {})
         for rel in data:
             if 'type' not in rel:
-                detail = 'Must specify correct data type'
-                return error_response(400, detail=detail)
+                return error_response(400, detail='Must specify correct data type')
             if 'id' not in rel:
-                detail = 'Must specify resource ID'
-                return error_response(400, detail=detail)
+                return error_response(400, detail='Must specify resource ID')
             type_ = rel['type']
             # The type name must match the collection name of model of the
             # relation.
-            if type_ != collection_name(related_model):
-                detail = ('Type must be {0}, not'
-                          ' {1}').format(collection_name(related_model), type_)
-                return error_response(409, detail=detail)
+            collection_name = self.api_manager.collection_name(related_model)
+            if type_ != collection_name:
+                return error_response(409, detail=f'Type must be {collection_name}, not {type_}')
             # Get the new objects to add to the relation.
-            new_value = get_by(self.session, related_model, rel['id'])
+            new_value = get_by(self.session, related_model, rel['id'], self.api_manager.primary_key_for(related_model))
             if new_value is None:
-                detail = ('No object of type {0} found with ID'
-                          ' {1}').format(type_, rel['id'])
-                return error_response(404, detail=detail)
+                return error_response(404, detail=f'No object of type {escape(type_)} found with ID {rel["id"]}')
             # Don't append a new value if it already exists in the to-many
             # relationship.
             if new_value not in related_value:
@@ -175,14 +155,13 @@ class RelationshipAPI(APIBase):
                     related_value.append(new_value)
                 except self.validation_exceptions as exception:
                     return self._handle_validation_exception(exception)
-        # TODO do we need to commit the session here?
-        #
-        #     self.session.commit()
-        #
+
+        self.session.commit()
+
         # Perform any necessary postprocessing.
         for postprocessor in self.postprocessors['POST_RELATIONSHIP']:
             postprocessor()
-        return {}, 204
+        return {}, 204, {}
 
     def patch(self, resource_id, relation_name):
         """Updates to a to-one or to-many relationship.
@@ -214,14 +193,10 @@ class RelationshipAPI(APIBase):
         # If no instance of the model exists with the specified instance ID,
         # return a 404 response.
         if instance is None:
-            detail = 'No instance with ID {0} in model {1}'
-            detail = detail.format(resource_id, self.model)
-            return error_response(404, detail=detail)
+            return error_response(404, detail=f'No instance with ID {escape(resource_id)} in model {self.model}')
         # If no such relation exists, return a 404.
         if not hasattr(instance, relation_name):
-            detail = 'Model {0} has no relation named {1}'
-            detail = detail.format(self.model, relation_name)
-            return error_response(404, detail=detail)
+            return error_response(404, detail=f'Model {self.model} has no relation named {escape(relation_name)}')
         related_model = get_related_model(self.model, relation_name)
         # related_value = getattr(instance, relation_name)
 
@@ -246,52 +221,43 @@ class RelationshipAPI(APIBase):
                 replacement = []
                 for rel in data:
                     if 'type' not in rel:
-                        detail = 'Must specify correct data type'
-                        return error_response(400, detail=detail)
+                        return error_response(400, detail='Must specify correct data type')
                     if 'id' not in rel:
-                        detail = 'Must specify resource ID or IDs'
-                        return error_response(400, detail=detail)
+                        return error_response(400, detail='Must specify resource ID or IDs')
                     type_ = rel['type']
                     # The type name must match the collection name of model of
                     # the relation.
-                    if type_ != collection_name(related_model):
-                        detail = 'Type must be {0}, not {1}'
-                        detail = detail.format(collection_name(related_model),
-                                               type_)
-                        return error_response(409, detail=detail)
+                    collection_name = self.api_manager.collection_name(related_model)
+                    if type_ != collection_name:
+                        return error_response(409, detail=f'Type must be {collection_name}, not {type_}')
                     id_ = rel['id']
-                    obj = get_by(self.session, related_model, id_)
+                    obj = get_by(self.session, related_model, id_, self.api_manager.primary_key_for(related_model))
                     replacement.append(obj)
             # Otherwise, we assume the client is trying to set a to-one
             # relationship.
             else:
                 if 'type' not in data:
-                    detail = 'Must specify correct data type'
-                    return error_response(400, detail=detail)
+                    return error_response(400, detail='Must specify correct data type')
                 if 'id' not in data:
-                    detail = 'Must specify resource ID or IDs'
-                    return error_response(400, detail=detail)
+                    return error_response(400, detail='Must specify resource ID or IDs')
                 type_ = data['type']
                 # The type name must match the collection name of model of the
                 # relation.
-                if type_ != collection_name(related_model):
-                    detail = ('Type must be {0}, not'
-                              ' {1}').format(collection_name(related_model),
-                                             type_)
-                    return error_response(409, detail=detail)
+                collection_name = self.api_manager.collection_name(related_model)
+                if type_ != collection_name:
+                    return error_response(409, detail=f'Type must be {collection_name}, not {type_}')
                 id_ = data['id']
-                replacement = get_by(self.session, related_model, id_)
+                replacement = get_by(self.session, related_model, id_, self.api_manager.primary_key_for(related_model))
             # If the to-one relationship resource or any of the to-many
             # relationship resources do not exist, return an error response.
             if replacement is None:
-                detail = ('No object of type {0} found'
-                          ' with ID {1}').format(type_, id_)
+                detail = f'No object of type {escape(type_)} found with ID {escape(id_)}'
                 return error_response(404, detail=detail)
             if isinstance(replacement, list) and any(value is None for value in replacement):
                 not_found = (rel for rel, value in zip(data, replacement)
                              if value is None)
                 detail = 'No object of type {0} found with ID {1}'
-                errors = [error(detail=detail.format(rel['type'], rel['id']))
+                errors = [error(detail=detail.format(escape(rel['type']), escape(rel['id'])))
                           for rel in not_found]
                 return errors_response(404, errors)
             # Finally, set the relationship to have the new value.
@@ -306,7 +272,7 @@ class RelationshipAPI(APIBase):
         # Perform any necessary postprocessing.
         for postprocessor in self.postprocessors['PATCH_RELATIONSHIP']:
             postprocessor()
-        return {}, 204
+        return {}, 204, {}
 
     def delete(self, resource_id, relation_name):
         """Deletes resources from a to-many relationship.
@@ -327,8 +293,7 @@ class RelationshipAPI(APIBase):
             data = json.loads(request.get_data()) or {}
         except (BadRequest, TypeError, ValueError, OverflowError) as exception:
             # this also happens when request.data is empty
-            detail = 'Unable to decode data'
-            return error_response(400, cause=exception, detail=detail)
+            return error_response(400, cause=exception, detail='Unable to decode data')
         was_deleted = False
         for preprocessor in self.preprocessors['DELETE_RELATIONSHIP']:
             temp_result = preprocessor(instance_id=resource_id,
@@ -340,11 +305,10 @@ class RelationshipAPI(APIBase):
                           self.primary_key)
         # If no such relation exists, return an error to the client.
         if not hasattr(instance, relation_name):
-            detail = 'No such link: {0}'.format(relation_name)
-            return error_response(404, detail=detail)
+            return error_response(404, detail=f'No such link: {escape(relation_name)}')
         # We assume that the relation is a to-many relation.
         related_model = get_related_model(self.model, relation_name)
-        related_type = collection_name(related_model)
+        related_type = self.api_manager.collection_name(related_model)
         relation = getattr(instance, relation_name)
         data = data.pop('data')
         not_found = []
@@ -359,18 +323,16 @@ class RelationshipAPI(APIBase):
             type_ = rel['type']
             id_ = rel['id']
             if type_ != related_type:
-                detail = ('Conflicting type: expected {0} but got type {1} for'
-                          ' linkage object with ID {2}')
-                detail = detail.format(related_type, type_, id_)
+                detail = f'Conflicting type: expected {related_type} but got type {escape(type_)} for linkage object with ID {escape(id_)}'
                 return error_response(409, detail=detail)
-            resource = get_by(self.session, related_model, id_)
+            resource = get_by(self.session, related_model, id_, self.api_manager.primary_key_for(related_model))
             if resource is None:
                 not_found.append((type_, id_))
             else:
                 to_remove.append(resource)
         if not_found:
             detail = 'No resource of type {0} and ID {1} found'
-            errors = [error(detail=detail.format(t, i)) for t, i in not_found]
+            errors = [error(detail=detail.format(escape(t), escape(i))) for t, i in not_found]
             return errors_response(404, errors)
         # Remove each of the resources from the relation (if they are not
         # already absent).
@@ -389,4 +351,4 @@ class RelationshipAPI(APIBase):
         if not was_deleted:
             detail = 'There was no instance to delete'
             return error_response(404, detail=detail)
-        return {}, 204
+        return {}, 204, {}
